require 'torch'   -- torch
require 'nn'      -- provides a normalization operator
require 'optim'
require 'image'

----------------------------------------------------------------------
-- Create a colormap
local colormap = torch.FloatTensor(8, 3)

--DAISY -> BLACK
colormap[1][1] = 0
colormap[1][2] = 0 
colormap[1][3] = 0
--ELAS -> RED
colormap[2][1] = 1
colormap[2][2] = 0 
colormap[2][3] = 0
--FCFV -> GREEN
colormap[3][1] = 0
colormap[3][2] = 1 
colormap[3][3] = 0
--MR-Cost -> BLUE
colormap[4][1] = 0
colormap[4][2] = 0 
colormap[4][3] = 1
--SH-SOB21 -> WHITE
colormap[5][1] = 1
colormap[5][2] = 1 
colormap[5][3] = 1
--SH-SSD5 -> YELLOW
colormap[6][1] = 1
colormap[6][2] = 1 
colormap[6][3] = 0
--SH-ZNCC21 -> CYAN
colormap[7][1] = 0
colormap[7][2] = 1 
colormap[7][3] = 1
--SUPER-rSGM5 -> PURPLE
colormap[8][1] = 1
colormap[8][2] = 0 
colormap[8][3] = 1
----------------------------------------------------------------------

-- conversion of a convolutional neural network to a fully convolutional model
function fully_connected_to_fully_convolutional(model)
	local fcnet = nn.Sequential()
	
	for i, v in ipairs(model.modules) do 
		if torch.type(v) == 'nn.Linear' then
			-- create fully convolutional layer
			local ninput = v.weight:size()[2]
			local noutput = v.weight:size()[1]
			local fclayer = nn.SpatialConvolution(ninput, noutput, 1, 1, 1, 1)
			fclayer.weight = torch.reshape(v.weight, ninput*noutput, 1, 1) 
			fclayer.bias = v.bias 
			fcnet:add(fclayer)
		else
			if torch.type(v) ~= 'nn.Reshape' then
				fcnet:add(v)	
			end
		end
	end
	return fcnet
end
----------------------------------------------------------------------

-- single, full resolution forward 
-- maximum speed (10 seconds on i7), maximum memory requirements (8 GB)
function singleForward()
	local image_id = '000'..opt.img

	if opt.img < 100
	then
		image_id = '0000'..opt.img	
	end

	if opt.img < 10
	then
		image_id = '00000'..opt.img	
	end

	local gt = image.load('/KITTI/disp_noc/'..image_id..'_10.png')
	local sample = torch.zeros(8, gt:size()[2], gt:size()[3])

	print(sys.COLORS.red..'Loading MSI tensor...')
	sample[1] = image.load(opt.inputdir..image_id..'_DAISY.pgm'):float()
	sample[2] = image.load(opt.inputdir..image_id..'_ELAS.pgm'):float()
	sample[3] = image.load(opt.inputdir..image_id..'_RHEM.pgm'):float()
	sample[4] = image.load(opt.inputdir..image_id..'_MRF-Cost.pgm'):float()
	sample[5] = image.load(opt.inputdir..image_id..'_SH-SOB21.pgm'):float()
	sample[6] = image.load(opt.inputdir..image_id..'_SH-SSD5.pgm'):float()
	sample[7] = image.load(opt.inputdir..image_id..'_SH-ZNCC21.pgm'):float()
	sample[8] = image.load(opt.inputdir..image_id..'_SUPER-rSGM5.pgm'):float()

	padder_left = nn.Padding(2, -4, 8, 0) 
	padder_right = nn.Padding(2, 4, 8, 0) 
	padder_up = nn.Padding(1, -4, 8, 0) 
	padder_down = nn.Padding(1, 4, 8, 0) 

	input = padder_left:forward(sample)
	input = padder_right:forward(input)
	input = padder_up:forward(input)
	input = padder_down:forward(input)

	output = model:forward(input)
	print(sys.COLORS.yellow..'Choice map tensor ready!')

	_, winners = torch.max(output, 1)
	local final_map = torch.zeros(1, gt:size()[2], gt:size()[3])
	local choosen_map = torch.zeros(3, gt:size()[2], gt:size()[3])

	local difference = 0
	local wrong_opt = 0
	local wrong_sgm = 0
	local wrong = 0
	local tot = 0
	for i=1,winners:size()[2] do
		for j=1,winners:size()[3] do

			-- assign disparity from the winner matcher
			final_map[1][i][j] = sample[winners[1][i][j]][i][j]

			-- apply colormap
			choosen_map[1][i][j] = colormap[winners[1][i][j]][1]
			choosen_map[2][i][j] = colormap[winners[1][i][j]][2]
			choosen_map[3][i][j] = colormap[winners[1][i][j]][3]		
		end
	end

	image.save(opt.confdir..image_id..'_DAISY.pgm', output[1])
	image.save(opt.confdir..image_id..'_ELAS.pgm', output[2])
	image.save(opt.confdir..image_id..'_RHEM.pgm', output[3])
	image.save(opt.confdir..image_id..'_MRF-Cost.pgm', output[4])
	image.save(opt.confdir..image_id..'_SH-SOB21.pgm', output[5])
	image.save(opt.confdir..image_id..'_SH-SSD5.pgm', output[6])
	image.save(opt.confdir..image_id..'_SH-ZNCC21.pgm', output[7])
	image.save(opt.confdir..image_id..'_SUPER-rSGM5.pgm', output[8])
	image.save(opt.stereodir..image_id..'.pgm', final_map)
	image.save(opt.stereodir..image_id..'.png', choosen_map)
	print(sys.COLORS.green..'Confidences, choice and final disparity maps saved!')
end
----------------------------------------------------------------------

-- double forward on crop images
-- lower memory requirements (4 GB)
function doubleForward()
	local image_id = '000'..opt.img

	if opt.img < 100
	then
		image_id = '0000'..opt.img	
	end

	if opt.img < 10
	then
		image_id = '00000'..opt.img	
	end

	local gt = image.load('/KITTI/disp_noc/'..image_id..'_10.png')
	local sample0 = torch.zeros(8, gt:size()[2], gt:size()[3])
	local sample1 = torch.zeros(8, gt:size()[2], gt:size()[3]/2)
	local sample2 = torch.zeros(8, gt:size()[2], gt:size()[3]/2)

	print(sys.COLORS.red..'Loading MSI tensor...')
	sample0[1] = image.load(opt.inputdir..image_id..'_DAISY.pgm'):float()
	sample0[2] = image.load(opt.inputdir..image_id..'_ELAS.pgm'):float()
	sample0[3] = image.load(opt.inputdir..image_id..'_RHEM.pgm'):float()
	sample0[4] = image.load(opt.inputdir..image_id..'_MRF-Cost.pgm'):float()
	sample0[5] = image.load(opt.inputdir..image_id..'_SH-SOB21.pgm'):float()
	sample0[6] = image.load(opt.inputdir..image_id..'_SH-SSD5.pgm'):float()
	sample0[7] = image.load(opt.inputdir..image_id..'_SH-ZNCC21.pgm'):float()
	sample0[8] = image.load(opt.inputdir..image_id..'_SUPER-rSGM5.pgm'):float()

	sample1[1] = image.load(opt.inputdir..image_id..'_DAISY.pgm'):float():narrow(3, 1, gt:size()[3]/2)
	sample1[2] = image.load(opt.inputdir..image_id..'_ELAS.pgm'):float():narrow(3, 1, gt:size()[3]/2)
	sample1[3] = image.load(opt.inputdir..image_id..'_RHEM.pgm'):float():narrow(3, 1, gt:size()[3]/2)
	sample1[4] = image.load(opt.inputdir..image_id..'_MRF-Cost.pgm'):float():narrow(3, 1, gt:size()[3]/2)
	sample1[5] = image.load(opt.inputdir..image_id..'_SH-SOB21.pgm'):float():narrow(3, 1, gt:size()[3]/2)
	sample1[6] = image.load(opt.inputdir..image_id..'_SH-SSD5.pgm'):float():narrow(3, 1, gt:size()[3]/2)
	sample1[7] = image.load(opt.inputdir..image_id..'_SH-ZNCC21.pgm'):float():narrow(3, 1, gt:size()[3]/2)
	sample1[8] = image.load(opt.inputdir..image_id..'_SUPER-rSGM5.pgm'):float():narrow(3, 1, gt:size()[3]/2)

	sample2[1] = image.load(opt.inputdir..image_id..'_DAISY.pgm'):float():narrow(3, gt:size()[3]/2, gt:size()[3]/2)
	sample2[2] = image.load(opt.inputdir..image_id..'_ELAS.pgm'):float():narrow(3, gt:size()[3]/2, gt:size()[3]/2)
	sample2[3] = image.load(opt.inputdir..image_id..'_RHEM.pgm'):float():narrow(3, gt:size()[3]/2, gt:size()[3]/2)
	sample2[4] = image.load(opt.inputdir..image_id..'_MRF-Cost.pgm'):narrow(3, gt:size()[3]/2, gt:size()[3]/2)
	sample2[5] = image.load(opt.inputdir..image_id..'_SH-SOB21.pgm'):narrow(3, gt:size()[3]/2, gt:size()[3]/2)
	sample2[6] = image.load(opt.inputdir..image_id..'_SH-SSD5.pgm'):narrow(3, gt:size()[3]/2, gt:size()[3]/2)
	sample2[7] = image.load(opt.inputdir..image_id..'_SH-ZNCC21.pgm'):narrow(3, gt:size()[3]/2, gt:size()[3]/2)
	sample2[8] = image.load(opt.inputdir..image_id..'_SUPER-rSGM5.pgm'):narrow(3, gt:size()[3]/2, gt:size()[3]/2)

	padder_left = nn.Padding(2, -4, 8, 0) 
	padder_right = nn.Padding(2, 4, 8, 0) 
	padder_up = nn.Padding(1, -4, 8, 0) 
	padder_down = nn.Padding(1, 4, 8, 0) 

	input1 = padder_left:forward(sample1)
	input1 = padder_right:forward(input1)
	input1 = padder_up:forward(input1)
	input1 = padder_down:forward(input1)

	padder_left = nn.Padding(2, -4, 8, 0) 
	padder_right = nn.Padding(2, 4, 8, 0) 
	padder_up = nn.Padding(1, -4, 8, 0) 
	padder_down = nn.Padding(1, 4, 8, 0) 

	input2 = padder_left:forward(sample2)
	input2 = padder_right:forward(input2)
	input2 = padder_up:forward(input2)
	input2 = padder_down:forward(input2)

	output1 = model:forward(input1):clone()
	output2 = model:forward(input2):clone()
	output = torch.cat(output1, output2, 3)
	print(sys.COLORS.yellow..'Choice map tensor ready!')

	_, winners = torch.max(output, 1)
	local final_map = torch.zeros(1, gt:size()[2], gt:size()[3])
	local choosen_map = torch.zeros(3, gt:size()[2], gt:size()[3])

	local difference = 0
	local wrong_opt = 0
	local wrong_sgm = 0
	local wrong = 0
	local tot = 0
	for i=1,winners:size()[2] do
		for j=1,winners:size()[3] do

			-- assign disparity from the winner matcher
			final_map[1][i][j] = sample0[winners[1][i][j]][i][j]

			-- apply colormap
			choosen_map[1][i][j] = colormap[winners[1][i][j]][1]
			choosen_map[2][i][j] = colormap[winners[1][i][j]][2]
			choosen_map[3][i][j] = colormap[winners[1][i][j]][3]		
		end
	end

	image.save(opt.confdir..image_id..'_DAISY.pgm', output[1])
	image.save(opt.confdir..image_id..'_ELAS.pgm', output[2])
	image.save(opt.confdir..image_id..'_RHEM.pgm', output[3])
	image.save(opt.confdir..image_id..'_MRF-Cost.pgm', output[4])
	image.save(opt.confdir..image_id..'_SH-SOB21.pgm', output[5])
	image.save(opt.confdir..image_id..'_SH-SSD5.pgm', output[6])
	image.save(opt.confdir..image_id..'_SH-ZNCC21.pgm', output[7])
	image.save(opt.confdir..image_id..'_SUPER-rSGM5.pgm', output[8])
	image.save(opt.stereodir..image_id..'.pgm', final_map)
	image.save(opt.stereodir..image_id..'.png', choosen_map)
	print(sys.COLORS.green..'Confidences, choice and final disparity maps saved!')
end
----------------------------------------------------------------------
-- parse command line arguments
if not opt then
   cmd = torch.CmdLine()
   cmd:text()
   cmd:text('Dataset Preprocessing')
   cmd:text()
   cmd:text('Options:')
   --cmd:option('-portion',1,'portion of images for the training set, a number in [0,1]')
   cmd:option('-img', 0, 'image to be tested')
   cmd:option('-model', 'DSF_3DV2016.net', 'network model')
   cmd:option('-log', './FUSIONS/error.csv', 'network model')
   cmd:option('-inputdir', './Input/', 'input directory')
   cmd:option('-stereodir', './Output/', 'stereo output directory')
   cmd:option('-confdir', './Output/', 'confidence output directory')
   cmd:option('-mode', 'single', 'single or double forward')
   cmd:text()
   opt = cmd:parse(arg or {})
end
----------------------------------------------------------------------

torch.setdefaulttensortype('torch.FloatTensor') -- preprocessing requires a floating point representation 

model = fully_connected_to_fully_convolutional(torch.load(opt.model, 'ascii'))
model:add(nn.Sigmoid())

if opt.mode == 'single'
then
	singleForward()
else
	doubleForward()
end
